/*
 * Copyright (C) 2016 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "Color.h"
#include "IntRect.h"
#include "IntSize.h"
#include "NativeImage.h"
#include "SharedBuffer.h"
//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#include "Logging.h"
//-TELESCOPE_LIB_CHANGES

namespace WebCore {

#if USE(CAIRO)
// Due to the pixman 16.16 floating point representation, cairo is not able to handle
// images whose size is bigger than 32768.
static const int cairoMaxImageSize = 32768;
#endif

class ImageBackingStore {
    WTF_MAKE_FAST_ALLOCATED;
public:
    static std::unique_ptr<ImageBackingStore> create(const IntSize& size, bool premultiplyAlpha = true)
    {
        return std::unique_ptr<ImageBackingStore>(new ImageBackingStore(size, premultiplyAlpha));
    }

    static std::unique_ptr<ImageBackingStore> create(const ImageBackingStore& other)
    {
        return std::unique_ptr<ImageBackingStore>(new ImageBackingStore(other));
    }

    NativeImagePtr image() const;

    bool setSize(const IntSize& size)
    {
        if (size.isEmpty())
            return false;

//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE( TSRENDER )
        if (!m_tsRenderBitmap || (m_tsRenderBitmap->Width() != size.width() || m_tsRenderBitmap->Height() != size.height()))
        {
            m_nativeImageptr.reset();
            m_nativeImageptr = TSRenderGlobal::getInstance().getRender()->CreateImage( size.width(), size.height(), tsRender::PixelFormat::BGRA8_UNORM );
            if ( !m_nativeImageptr )
            {
                RELEASE_LOG_INFO( Images, "ImageBackingStore create image failed: (%d, %d)", size.width(), size.height() );
                return false;
            }

            m_tsRenderBitmap = tsRender::Bitmap::CreateUpload(size.width(), size.height(), tsRender::PixelFormat::BGRA8_UNORM);
		    if (!m_tsRenderBitmap)
		    {
                m_nativeImageptr.reset();
			    RELEASE_LOG_INFO( Images, "ImageBackingStore create upload bitmap failed: (%d, %d)", size.width(), size.height() );
			    return false;
		    }
        }
#else
        Vector<char> buffer;
#if PLATFORM(TELESCOPE)
        buffer.setMemoryTag(&WTF_IMAGE_MEMORY_TAG);
#endif // #if PLATFORM(TELESCOPE)
        size_t bufferSize = size.area().unsafeGet() * sizeof(RGBA32);

        if (!buffer.tryReserveCapacity(bufferSize))
            return false;

        buffer.grow(bufferSize);
        m_pixels = SharedBuffer::DataSegment::create(WTFMove(buffer));
        m_pixelsPtr = reinterpret_cast<RGBA32*>(const_cast<char*>(m_pixels->data()));
#endif
//-TELESCOPE_LIB_CHANGES
        m_size = size;
        m_frameRect = IntRect(IntPoint(), m_size);
        clear();
        return true;
    }

    void setFrameRect(const IntRect& frameRect)
    {
        ASSERT(!m_size.isEmpty());
        ASSERT(inBounds(frameRect));
        m_frameRect = frameRect;
    }

    const IntSize& size() const { return m_size; }
    const IntRect& frameRect() const { return m_frameRect; }

//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE(TSRENDER)
	void *GetTSRenderBitmapBuffer() const
	{
		if (!m_tsRenderBitmap || !m_tsRenderBitmap->Data())
		{
			return nullptr;
		}
		return m_tsRenderBitmap->Data();
	}
#endif
//-TELESCOPE_LIB_CHANGES
	
    void clear()
    {
//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE(TSRENDER)
        void *pBuf = GetTSRenderBitmapBuffer();
		if (!pBuf)
		{
			RELEASE_LOG_INFO( Telescope, "m_tsRenderBitmap data is empty: %s:%d", __FILE__, __LINE__ );
			return;
		}
		memset(pBuf, 0, m_tsRenderBitmap->Pitch() * m_tsRenderBitmap->Height());
#else
        memset(m_pixelsPtr, 0, (m_size.area() * sizeof(RGBA32)).unsafeGet());
#endif
//-TELESCOPE_LIB_CHANGES
    }

    void clearRect(const IntRect& rect)
    {
        if (rect.isEmpty() || !inBounds(rect))
            return;

        size_t rowBytes = rect.width() * sizeof(RGBA32);
        RGBA32* start = pixelAt(rect.x(), rect.y());
        //+TELESCOPE_LIB_CHANGES
        //5/23/2023
        //Use move instead of create a new image for image backing store
        if (start)
        {
            for (int i = 0; i < rect.height(); ++i) {
                memset(start, 0, rowBytes);
                start += m_size.width();
            }
        }
        //-TELESCOPE_LIB_CHANGES
    }

    void fillRect(const IntRect &rect, unsigned r, unsigned g, unsigned b, unsigned a)
    {
        if (rect.isEmpty() || !inBounds(rect))
            return;

        //+TELESCOPE_LIB_CHANGES
        //5/23/2023
        //Use move instead of create a new image for image backing store
        RGBA32* start = pixelAt(rect.x(), rect.y());
        if (start)
        {
            RGBA32 pixelValue = this->pixelValue(r, g, b, a);
            for (int i = 0; i < rect.height(); ++i) {
                for (int j = 0; j < rect.width(); ++j)
                    start[j] = pixelValue;
                start += m_size.width();
            }
        }
        //-TELESCOPE_LIB_CHANGES
    }

    void repeatFirstRow(const IntRect& rect)
    {
        if (rect.isEmpty() || !inBounds(rect))
            return;

        size_t rowBytes = rect.width() * sizeof(RGBA32);
        RGBA32* src = pixelAt(rect.x(), rect.y());
        //+TELESCOPE_LIB_CHANGES
        //5/23/2023
        //Use move instead of create a new image for image backing store
        if (src)
        {
            RGBA32* dest = src + m_size.width();
            for (int i = 1; i < rect.height(); ++i) {
                memcpy(dest, src, rowBytes);
                dest += m_size.width();
            }
        }
        //-TELESCOPE_LIB_CHANGES
    }

    RGBA32* pixelAt(int x, int y) const
    {
        ASSERT(inBounds(IntPoint(x, y)));
//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE(TSRENDER)
		void *pBuf = GetTSRenderBitmapBuffer();
		if (!pBuf)
		{
			RELEASE_LOG_INFO( Telescope, "m_tsRenderBitmap data is empty: %s:%d", __FILE__, __LINE__ );
			return nullptr;
		}
        return static_cast<RGBA32*>(pBuf) + y * m_size.width() + x;
#else
        return m_pixelsPtr + y * m_size.width() + x;
#endif
//-TELESCOPE_LIB_CHANGES
    }

    void setPixel(RGBA32* dest, unsigned r, unsigned g, unsigned b, unsigned a)
    {
        if (!dest)
        {
            RELEASE_LOG_INFO( Telescope, "setPixel dest is nullptr: %s:%d", __FILE__, __LINE__ );
            return;
        }
        *dest = pixelValue(r, g, b, a);
    }

    void setPixel(int x, int y, unsigned r, unsigned g, unsigned b, unsigned a)
    {
        RGBA32* pPixel = pixelAt(x, y);
        //+TELESCOPE_LIB_CHANGES
        //5/23/2023
        //Use move instead of create a new image for image backing store
        if (pPixel)
        {
            setPixel(pPixel, r, g, b, a);
        }
        //-TELESCOPE_LIB_CHANGES
    }

    void blendPixel(RGBA32* dest, unsigned r, unsigned g, unsigned b, unsigned a)
    {
        if (!a)
            return;

        if (a >= 255 || !alphaChannel(*dest)) {
            setPixel(dest, r, g, b, a);
            return;
        }

        if (!m_premultiplyAlpha)
            *dest = makePremultipliedRGBA(redChannel(*dest), greenChannel(*dest), blueChannel(*dest), alphaChannel(*dest), false);

        unsigned d = 255 - a;

        r = fastDivideBy255(r * a + redChannel(*dest) * d);
        g = fastDivideBy255(g * a + greenChannel(*dest) * d);
        b = fastDivideBy255(b * a + blueChannel(*dest) * d);
        a += fastDivideBy255(d * alphaChannel(*dest));

        if (m_premultiplyAlpha)
            *dest = makeRGBA(r, g, b, a);
        else
            *dest = makeUnPremultipliedRGBA(r, g, b, a);
    }

    static bool isOverSize(const IntSize& size)
    {
#if USE(CAIRO)
        // FIXME: this is a workaround to avoid the cairo image size limit, but we should implement support for
        // bigger images. See https://bugs.webkit.org/show_bug.cgi?id=177227.
        //
        // If the image is bigger than the cairo limit it can't be displayed, so we don't even try to decode it.
        if (size.width() > cairoMaxImageSize || size.height() > cairoMaxImageSize)
            return true;
#endif
        static unsigned long long MaxPixels = ((1 << 29) - 1);
        unsigned long long pixels = static_cast<unsigned long long>(size.width()) * static_cast<unsigned long long>(size.height());
        return pixels > MaxPixels;
    }

private:
    ImageBackingStore(const IntSize& size, bool premultiplyAlpha = true)
        : m_premultiplyAlpha(premultiplyAlpha)
    {
        ASSERT(!size.isEmpty() && !isOverSize(size));
        setSize(size);
    }

    ImageBackingStore(const ImageBackingStore& other)
        : m_size(other.m_size)
        , m_premultiplyAlpha(other.m_premultiplyAlpha)
    {
        ASSERT(!m_size.isEmpty() && !isOverSize(m_size));

//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE(TSRENDER)
        setSize(m_size);
#else
        Vector<char> buffer;
        buffer.append(other.m_pixels->data(), other.m_pixels->size());
        m_pixels = SharedBuffer::DataSegment::create(WTFMove(buffer));
        m_pixelsPtr = reinterpret_cast<RGBA32*>(const_cast<char*>(m_pixels->data()));
#endif
//-TELESCOPE_LIB_CHANGES
    }

    bool inBounds(const IntPoint& point) const
    {
        return IntRect(IntPoint(), m_size).contains(point);
    }

    bool inBounds(const IntRect& rect) const
    {
        return IntRect(IntPoint(), m_size).contains(rect);
    }

    RGBA32 pixelValue(unsigned r, unsigned g, unsigned b, unsigned a) const
    {
        if (m_premultiplyAlpha && !a)
            return 0;

        if (m_premultiplyAlpha && a < 255)
            return makePremultipliedRGBA(r, g, b, a, false);

        return makeRGBA(r, g, b, a);
    }

//+TELESCOPE_LIB_CHANGES
//5/23/2023
//Use move instead of create a new image for image backing store
#if USE(TSRENDER)
    mutable tsRender::RefPtr<tsRender::Bitmap> m_tsRenderBitmap;
    mutable NativeImagePtr m_nativeImageptr;
#else
    RefPtr<SharedBuffer::DataSegment> m_pixels;
    RGBA32* m_pixelsPtr { nullptr };
#endif
//-TELESCOPE_LIB_CHANGES

    IntSize m_size;
    IntRect m_frameRect; // This will always just be the entire buffer except for GIF and PNG frames
    bool m_premultiplyAlpha { true };
};

}
