/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ImageBuffer.h"

#if USE( TSRENDER )
#include "GraphicsContextImpl.h"
#include <wtf/text/Base64.h>
#include "NotImplemented.h"
#include "TSRender.h"
#include "CanvasImage.h"

namespace WebCore
{

ImageBufferData::ImageBufferData( const IntSize &size, RenderingMode renderingMode )
	: m_size( size )
	, m_canvasImage( CanvasImage::create( size, false ) )
{
}


ImageBufferData::~ImageBufferData()
{
}


RefPtr<JSC::Uint8ClampedArray> ImageBufferData::getData( AlphaPremultiplication outputFormat, const IntRect &rect, const IntSize &size ) const
{
	Checked<unsigned, RecordOverflow> area = 4;
	return Uint8ClampedArray::tryCreateUninitialized( area.unsafeGet() );
}


void ImageBufferData::putData( const Uint8ClampedArray &source, AlphaPremultiplication sourceFormat, const IntSize &sourceSize, const IntRect &sourceRect, const IntPoint &destPoint, const IntSize &size )
{
	notImplemented();
}


Vector<uint8_t> ImageBuffer::toBGRAData() const
{
	notImplemented();
	return Vector<uint8_t>();
}


ImageBuffer::ImageBuffer( const FloatSize &size, float resolutionScale, ColorSpace, RenderingMode renderingMode, const HostWindow *, bool &success )
	: m_data( IntSize( size ), renderingMode )
	, m_size( size )
	, m_logicalSize( size )
	, m_resolutionScale( resolutionScale )
{
	notImplemented();
	success = false;
	if ( m_size.isEmpty() )
	{
		return;
	}
}

ImageBuffer::~ImageBuffer() = default;

std::unique_ptr<ImageBuffer> ImageBuffer::createCompatibleBuffer( const FloatSize &size, const GraphicsContext &context )
{
	return createCompatibleBuffer( size, ColorSpace::SRGB, context );
}


GraphicsContext &ImageBuffer::context() const
{
	notImplemented();
	assert( false );
	GraphicsContext *pDummy = nullptr;
	return *pDummy;
}


RefPtr<Image> ImageBuffer::sinkIntoImage( std::unique_ptr<ImageBuffer> imageBuffer, PreserveResolution preserveResolution )
{
	return imageBuffer->copyImage( DontCopyBackingStore, preserveResolution );
}


RefPtr<Image> ImageBuffer::copyImage( BackingStoreCopy copyBehavior, PreserveResolution ) const
{
	return m_data.m_canvasImage;
}


void ImageBuffer::drawConsuming( std::unique_ptr<ImageBuffer> imageBuffer, GraphicsContext &destContext, const FloatRect &destRect, const FloatRect &srcRect, const ImagePaintingOptions &options )
{
	imageBuffer->draw( destContext, destRect, srcRect, options );
}


void ImageBuffer::draw( GraphicsContext &context, const FloatRect &destRect, const FloatRect &srcRect, const ImagePaintingOptions &options )
{
	m_data.m_canvasImage->draw( context, destRect, srcRect, options );
}


void ImageBuffer::drawPattern( GraphicsContext &context, const FloatRect &destRect, const FloatRect &srcRect, const AffineTransform &patternTransform,
	const FloatPoint &phase, const FloatSize &spacing, const ImagePaintingOptions &options )
{
	m_data.m_canvasImage->drawPattern( context, destRect, srcRect, patternTransform, phase, spacing, options );
}


void ImageBuffer::platformTransformColorSpace( const std::array<uint8_t, 256> &lookUpTable )
{
	notImplemented();
}


RefPtr<Uint8ClampedArray> ImageBuffer::getUnmultipliedImageData( const IntRect &rect, IntSize *pixelArrayDimensions, CoordinateSystem coordinateSystem ) const
{
	IntRect srcRect = rect;
	if ( coordinateSystem == LogicalCoordinateSystem )
	{
		srcRect.scale( m_resolutionScale );
	}

	if ( pixelArrayDimensions )
	{
		*pixelArrayDimensions = srcRect.size();
	}
	return m_data.getData( AlphaPremultiplication::Unpremultiplied, srcRect, internalSize() );
}


RefPtr<Uint8ClampedArray> ImageBuffer::getPremultipliedImageData( const IntRect &rect, IntSize *pixelArrayDimensions, CoordinateSystem coordinateSystem ) const
{
	IntRect srcRect = rect;
	if ( coordinateSystem == LogicalCoordinateSystem )
	{
		srcRect.scale( m_resolutionScale );
	}

	if ( pixelArrayDimensions )
	{
		*pixelArrayDimensions = srcRect.size();
	}
	return m_data.getData( AlphaPremultiplication::Premultiplied, srcRect, internalSize() );
}


void ImageBuffer::putByteArray( const Uint8ClampedArray &source, AlphaPremultiplication sourceFormat, const IntSize &sourceSize, const IntRect &sourceRect, const IntPoint &destPoint, CoordinateSystem coordinateSystem )
{
	IntRect scaledSourceRect = sourceRect;
	IntSize scaledSourceSize = sourceSize;
	if ( coordinateSystem == LogicalCoordinateSystem )
	{
		scaledSourceRect.scale( m_resolutionScale );
		scaledSourceSize.scale( m_resolutionScale );
	}
	m_data.putData( source, sourceFormat, scaledSourceSize, scaledSourceRect, destPoint, internalSize() );
}


String ImageBuffer::toDataURL( const String &mimeType, Optional<double> quality, PreserveResolution ) const
{
	Vector<uint8_t> encodedImage = toData( mimeType, quality );
	if ( encodedImage.isEmpty() )
	{
		return "data:,";
	}

	Vector<char> base64Data;
	base64Encode( encodedImage.data(), encodedImage.size(), base64Data );

	return "data:" + mimeType + ";base64," + base64Data;
}


Vector<uint8_t> ImageBuffer::toData( const String &mimeType, Optional<double> quality ) const
{
	notImplemented();
	return Vector<uint8_t>();
}

} // namespace WebCore

#endif // #if USE( TSRENDER )
