/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>
#include "GraphicsContext.h"
#include "NativeImage.h"
#include "GraphicsTypes.h"
#include "PlatformContextTSRender.h"
#include "TSRender.h"
#include "NotImplemented.h"

#if USE( TSRENDER )
namespace WebCore
{

IntSize nativeImageSize( const NativeImagePtr &image )
{
	if ( !image )
	{
		return IntSize();
	}
	return IntSize( image->Width(), image->Height() );
}


bool nativeImageHasAlpha( const NativeImagePtr &image )
{
	notImplemented();
	return true;
}


Color nativeImageSinglePixelSolidColor( const NativeImagePtr &image )
{
	notImplemented();
	return Color();
}


void drawNativeImage( const NativeImagePtr &image, GraphicsContext &context, const FloatRect &destRect
	, const FloatRect &srcRect, const IntSize &imageSize, const ImagePaintingOptions &options )
{
	auto orientation = options.orientation();

	FloatRect adjustedSrcRect{srcRect};
	FloatRect adjustedDestRect{destRect};
	bool needTransform = orientation != ImageOrientation::None;
	if ( needTransform )
	{
		FloatSize adjustedImageSize{ imageSize };
		adjustedDestRect.setLocation( FloatPoint() );
		if ( orientation.usesWidthAsHeight() )
		{
			adjustedImageSize = imageSize.transposedSize();
			adjustedDestRect.setSize(destRect.size().transposedSize());
		}

		switch( orientation )
		{
		case ImageOrientation::FromImage:
			ASSERT_NOT_REACHED();
			break;
		case ImageOrientation::OriginTopLeft:
			break;
		case ImageOrientation::OriginTopRight:
			// Flip along y axis
			adjustedSrcRect.setX( adjustedImageSize.width() - srcRect.width() - srcRect.x() );
			break;
		case ImageOrientation::OriginBottomRight:
			// 180 degree rotation
			adjustedSrcRect.setX( adjustedImageSize.width() - srcRect.width() - srcRect.x() );
			adjustedSrcRect.setY( adjustedImageSize.height() - srcRect.height() - srcRect.y() );
			break;
		case ImageOrientation::OriginBottomLeft:
			// Flip along x axis
			adjustedSrcRect.setY( adjustedImageSize.height() - srcRect.height() - srcRect.y() );
			break;
		case ImageOrientation::OriginLeftTop:
			// mirror along x-axis + 270 degree CW rotation
			adjustedSrcRect.setX( adjustedImageSize.width() - srcRect.height() - srcRect.y() );
			adjustedSrcRect.setY( adjustedImageSize.height() - srcRect.width() - srcRect.x() );
			adjustedSrcRect.setWidth( srcRect.height() );
			adjustedSrcRect.setHeight( srcRect.width() );
			break;
		case ImageOrientation::OriginRightTop:
			// 90 degree CW rotation
			adjustedSrcRect.setX( srcRect.y() );
			adjustedSrcRect.setY( adjustedImageSize.height() - srcRect.width() - srcRect.x() );
			adjustedSrcRect.setWidth( srcRect.height() );
			adjustedSrcRect.setHeight( srcRect.width() );
			break;
		case ImageOrientation::OriginRightBottom:
			// mirror along x-axis + 90 degree CW rotation
			adjustedSrcRect.setX( srcRect.y() );
			adjustedSrcRect.setY( srcRect.x() );
			adjustedSrcRect.setWidth( srcRect.height() );
			adjustedSrcRect.setHeight( srcRect.width() );
			break;
		case ImageOrientation::OriginLeftBottom:
			// 270 degree CW rotation
			adjustedSrcRect.setX( adjustedImageSize.width() - srcRect.height() - srcRect.y() );
			adjustedSrcRect.setY( srcRect.x() );
			adjustedSrcRect.setWidth( srcRect.height() );
			adjustedSrcRect.setHeight( srcRect.width() );
			break;
		default:
			ASSERT_NOT_REACHED();
			break;
		}

		context.save();
		context.translate( destRect.location() );
		context.concatCTM( orientation.transformFromDefault( destRect.size() ) );
	}


	auto path = tsRender::Path::Create();
	path->Rect( ConvertFloatRectToTSRender( adjustedDestRect ) );
	auto paint = tsRender::ImagePaint::Create( ConvertFloatRectToTSRender( adjustedDestRect ), ConvertFloatRectToTSRender( adjustedSrcRect ), image, tsRender::ColorFromRGBA( 255, 255, 255, 255 ) );

	ASSERT( context.hasPlatformContext() );
	context.platformContext()->getSceneBuilder()->FillPath( *path, *paint );

	if( needTransform )
	{
		context.restore();
	}
}


void clearNativeImageSubimages( const NativeImagePtr &image )
{
	notImplemented();
}
} // namespace WebCore
#endif // #if USE( TSRENDER )