/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "WebFrame.h"

#include "MemoryTrackingScope.h"

#include "JavaScriptCore/APICast.h"
#include "JavaScriptCore/JSGlobalObject.h"
#include "JavaScriptCore/PropertyName.h"
#include "JavaScript/TelescopeInstance.h"
#include "WebCore/Document.h"
#include "WebCore/DocumentLoader.h"
#include "WebCore/DOMWrapperWorld.h"
#include "webcore/Frame.h"
#include "webcore/FrameLoader.h"
#include "WebCore/FrameLoadRequest.h"
#include "WebCore/JSDOMWindowBase.h"
#include "webcore/Page.h"
#include "WebCore/RenderTreeAsText.h"
#include "webcore/ResourceRequest.h"
#include "WebCore/ResourceResponse.h"
#include "WebCore/ScriptController.h"
#include "WebFrameLoaderClient.h"
#include "WebView.h"

using namespace WebCore;
using namespace Telescope;

#define JSBINDING_BASIC_CHECK()                                                                \
	WebCore::JSDOMWindow *pWindow = toJSDOMWindow( mFrame, WebCore::mainThreadNormalWorld() ); \
	JSC::Bindings::RootObject *pRoot = mFrame->script().bindingRootObject();                   \
	if ( !pWindow || !pRoot )                                                                  \
	{                                                                                          \
		assert( 0 );                                                                           \
	}                                                                                          \
	JSC_GlobalObject *pGlobalObj = pWindow->globalObject();                                    \
	JSC::JSLockHolder lock( pGlobalObj );

WebCore::Frame *core( WebFrame *webFrame )
{
	return webFrame->frameView();
}


WebFrame::WebFrame()
	: mFrame( nullptr )
	, mWebView( nullptr )
{
}


WebFrame::~WebFrame()
{
}


Ref<WebCore::Frame> WebFrame::createSubframeWithOwnerElement( WebView *view, WebCore::Page *page, WebCore::HTMLFrameOwnerElement *ownerElement )
{
	mWebView = view;
	auto frame = Frame::create( page, ownerElement, new WebFrameLoaderClient( this ) );
	mFrame = frame.ptr();
	return frame;
}


void WebFrame::initWithWebView( ::Telescope::WebView *view, WebCore::Page *page )
{
	mWebView = view;
	mFrame = &page->mainFrame();
}


WebView *WebFrame::webView() const
{
	return mWebView;
}


WebFrame *WebFrame::createInstance()
{
	WebFrame *instance = new WebFrame();
	return instance;
}


void WebFrame::loadRequest( WTF::URL *url )
{
	TELESCOPE_MEMORY_SCOPE( "WebFrame::loadRequest" );

	Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return;
	}
	coreFrame->loader().load( FrameLoadRequest( *coreFrame, ResourceRequest( *url ), WebCore::ShouldOpenExternalURLsPolicy::ShouldAllow ) );
}


void WebFrame::AddToJavaScriptWindowObject( const char *pName, Telescope::IJSBoundObject *pObj )
{
	JSBINDING_BASIC_CHECK();

	JSC_Object *pRuntimeObj = JSC::Bindings::TelescopeInstance::Create( pObj, pRoot )->createRuntimeObject( pGlobalObj );
	JSC::Identifier oIdt = JSC::Identifier::fromString( pGlobalObj->vm(), pName );

	JSC::PutPropertySlot oSlot( pWindow );

	JSC::PropertyName oPropertyName( oIdt );
	pWindow->methodTable( pGlobalObj->vm() )->put( pWindow, pGlobalObj, oPropertyName, pRuntimeObj, oSlot );
}


Telescope::TelescopeJSValue *WebFrame::CreateJavaScriptValue()
{
	JSBINDING_BASIC_CHECK();

	JSC_Value newValue = JSC::jsUndefined();
	TelescopeJSValue *v = new TelescopeJSValue( &newValue, pGlobalObj );
	return v;
}


void WebFrame::DestroyJavaScriptValue( Telescope::TelescopeJSValue *value )
{
	JSBINDING_BASIC_CHECK();

	delete value;
}


WebCore::Document *WebFrame::DOMDocument()
{
	WebCore::Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return nullptr;
	}
	return coreFrame->document();
}


WebCore::Frame *WebFrame::parentFrame()
{
	WebCore::Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return nullptr;
	}
	return coreFrame->tree().parent();
}


void WebFrame::childFrames( Vector<WebFrame *>* vector )
{
	WebCore::Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return;
	}

	size_t size = coreFrame->tree().childCount();
	if ( !size )
	{
		return;
	}

	vector->reserveInitialCapacity( size );

	for ( WebCore::Frame *child = coreFrame->tree().firstChild(); child; child = child->tree().nextSibling() )
	{
		WebFrame *webFrame = WebFrame::fromCoreFrame( *child );
		ASSERT( webFrame );
		vector->uncheckedAppend( webFrame );
	}
}


WebFrame *WebFrame::fromCoreFrame( const WebCore::Frame &frame )
{
	auto *webFrameLoaderClient = toWebFrameLoaderClient( frame.loader().client() );
	if ( !webFrameLoaderClient )
	{
		return nullptr;
	}

	return webFrameLoaderClient->webFrame();
}


const WebCore::ResourceResponse *WebFrame::resourceResponse()
{
	WebCore::Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return nullptr;
	}
	DocumentLoader *documentLoader = coreFrame->loader().documentLoader();
	if ( !documentLoader )
	{
		return nullptr;
	}
	return &documentLoader->response();
}


JSGlobalContextRef WebFrame::globalContext()
{
	Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return nullptr;
	}

	return toGlobalRef( coreFrame->script().globalObject( mainThreadNormalWorld() ) );
}


void WebFrame::name(WTF::String* name)
{
	WebCore::Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return;
	}
	*name = coreFrame->tree().uniqueName();
}


enum WebRenderTreeAsTextOption
{
	WebRenderTreeAsTextShowAllLayers = 1 << 0,
	WebRenderTreeAsTextShowLayerNesting = 1 << 1,
	WebRenderTreeAsTextShowCompositedLayers = 1 << 2,
	WebRenderTreeAsTextShowOverflow = 1 << 3,
	WebRenderTreeAsTextShowSVGGeometry = 1 << 4,
	WebRenderTreeAsTextShowLayerFragments = 1 << 5
};


typedef unsigned WebRenderTreeAsTextOptions;


static OptionSet<RenderAsTextFlag> toRenderAsTextFlags( WebRenderTreeAsTextOptions options )
{
	OptionSet<RenderAsTextFlag> flags;

	if ( options & WebRenderTreeAsTextShowAllLayers )
	{
		flags.add( RenderAsTextFlag::ShowAllLayers );
	}
	if ( options & WebRenderTreeAsTextShowLayerNesting )
	{
		flags.add( RenderAsTextFlag::ShowLayerNesting );
	}
	if ( options & WebRenderTreeAsTextShowCompositedLayers )
	{
		flags.add( RenderAsTextFlag::ShowCompositedLayers );
	}
	if ( options & WebRenderTreeAsTextShowOverflow )
	{
		flags.add( RenderAsTextFlag::ShowOverflow );
	}
	if ( options & WebRenderTreeAsTextShowSVGGeometry )
	{
		flags.add( RenderAsTextFlag::ShowSVGGeometry );
	}
	if ( options & WebRenderTreeAsTextShowLayerFragments )
	{
		flags.add( RenderAsTextFlag::ShowLayerFragments );
	}

	return flags;
}


bool WebFrame::renderTreeAsExternalRepresentation( unsigned options, WTF::String *result )
{
	if ( !result )
	{
		return false;
	}

	Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return false;
	}

	*result = externalRepresentation( coreFrame, toRenderAsTextFlags( options ) );
	return true;
}


bool WebFrame::renderTreeAsExternalRepresentationForPrinting( WTF::String *result )
{
	if ( !result )
	{
		return false;
	}

	Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return false;
	}

	*result = externalRepresentation( coreFrame, { RenderAsTextFlag::PrintingMode } );
	return true;
}


unsigned WebFrame::pendingFrameUnloadEventCount()
{
	Frame *coreFrame = core( this );
	if ( !coreFrame )
	{
		return 0;
	}

	Document *document = coreFrame->document();
	if ( !document )
	{
		return 0;
	}

	DOMWindow *domWindow = document->domWindow();

	if ( !domWindow )
	{
		return 0;
	}

	return domWindow->pendingUnloadEventListeners();
}