/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if TELESCOPE_USING( TELESCOPE_TESTRUNNER )

#include "WebFrame.h"
#include "WebDumper.h"
#include "WebView.h"
#include "PixelDumpSupport.h"
#include "Tool/TestRunner/TestRunner.h"
#include <WebCore/ResourceResponse.h>
#include <WebCore/Element.h>

namespace Telescope 
{

static void dumpFramesAsText(WebFrame* frame, bool dumpChildFrame, String* oStr)
{
    if (!frame)
    {
        return;
    }

    WebCore::Document* document = frame->DOMDocument();
    if (!document)
    {
        return;
    }

    WebCore::Element* documentElement = document->documentElement();
    if (!documentElement)
    {
        return;
    }

    WebCore::Frame* parentFrame = frame->parentFrame();

    if (parentFrame) 
    {
        WTF::String parentFrameName;
        frame->name( &parentFrameName );
        oStr->append("\n--------\nFrame: '");
        oStr->append(parentFrameName);
        oStr->append("'\n--------\n");
    }

    oStr->append(documentElement->innerText());
    oStr->append("\n");

    if (dumpChildFrame) 
    {
        Vector<WebFrame*> webFrameVector;
        frame->childFrames(&webFrameVector);
        for (auto& webFrame : webFrameVector) {
            dumpFramesAsText(webFrame, dumpChildFrame, oStr);
        }
    }
}


void WebDumper::dumpRenderTree(WebFrame* frame, bool dumpAsText, bool dumpChildFramesAsText, bool isForPrinting, unsigned renderTreeOptions, bool toPrintSeparators)
{
    const WebCore::ResourceResponse* resourceResponse = frame->resourceResponse();

    if (resourceResponse)
    {
        const WTF::String& mimeType = resourceResponse->mimeType();
        if (mimeType == "text/plain")
        {
            dumpAsText = true;
        }
    }

    WTF::String renderTreeRepresentation;

    if (dumpAsText)
    {
        dumpFramesAsText(frame, dumpChildFramesAsText, &renderTreeRepresentation);
    }
    else 
    {
        if (isForPrinting)
        {
            frame->renderTreeAsExternalRepresentationForPrinting(&renderTreeRepresentation);
        }
        else
        {
            frame->renderTreeAsExternalRepresentation(renderTreeOptions, &renderTreeRepresentation);
        }
    }
            
    unsigned stringLength = renderTreeRepresentation.length();
    if (stringLength) 
    {
        TESTRUNNER_OUTPUT("%s", renderTreeRepresentation.utf8().data());
    }
    else
    {
        TESTRUNNER_OUTPUT("ERROR: nil result from %s\n", dumpAsText ? "IDOMElement::innerText" : "IFrameViewPrivate::renderTreeAsExternalRepresentation");
    }

    if (toPrintSeparators)
    {
        TESTRUNNER_OUTPUT("#EOF\n");
    }
}


void WebDumper::dumpPixelResultAndCompareWithExpected(WebFrame* frame)
{
    WebView* webView = frame->webView();
    webView->Paint();
    dumpWebViewAsPixelsAndCompareWithExpected(webView, gTestRunner->expectedPixelHash());
}

} // namespace Telescope

#endif // #if TELESCOPE_USING( TELESCOPE_TESTRUNNER )